; 
; Publics needed for the below procedures
;
; capal macro
;
; Constants
;     Terminator1  dw 0A0Ah
;     Terminator2  dw 0000h
;     Terminator3  dw 0909h
;     Terminator4  db 0Ah
;     Terminator5  db 00h
;     Terminator6  db 09h
;
; Variables
;     ptrTree      dd 0
;     ptrAddresses dd 0
;     ptrStrings   dd 0
;     ptrNumbs     dd 0
;
; PROTOs
;     MakeIndex       PROTO :DWORD, :DWORD
;     MakeIndex1      PROTO :DWORD, :DWORD
;     MakeIndex2      PROTO :DWORD, :DWORD
;     MakeIndex4      PROTO :DWORD, :DWORD
;
;     IndexedLookup   PROTO :DWORD, :DWORD, :DWORD, :NEAR, :WORD
;     IndexedLookup1  PROTO :DWORD, :DWORD, :DWORD, :NEAR, :WORD
;     IndexedLookup2  PROTO :DWORD, :DWORD, :DWORD, :NEAR, :WORD
;     IndexedLookup4  PROTO :DWORD, :DWORD, :DWORD, :NEAR, :WORD
;
;     line_cmp        PROTO :DWORD, :DWORD, :WORD
;     line_cmp_nocase PROTO :DWORD, :DWORD, :WORD
;     line_cmp_numb1  PROTO :DWORD, :DWORD, :WORD
;     line_cmp_numb2  PROTO :DWORD, :DWORD, :WORD
;     line_cmp_numb4  PROTO :DWORD, :DWORD, :WORD
;
;     SetStringAddresses PROTO :DWORD, :DWORD, :DWORD, :BYTE
;
;  invoking MakeIndex automatically invokes MakeIndex1, MakeIndex2, or MakeIndex4 pending nQty
;           invoke MakeIndex, ptrTree, nQty
;
;  invoking IndexedLookup automatically invokes IndexedLookup1, IndexedLookup2, or IndexedLookup4 pending nQty in Tree
;           invoke IndexedLookup, ADDR SB_0017, ptrTree, ptrAddresses, ADDR line_cmp, Terminator6
;
;  invoke SetStringAddresses, ptrAddresses, ptrStrings, nQty, Terminator3
;
;  Terminators 1 thru 3 are for IndexLooup because of word size format
;  Terminators 4 thru 6 are for SetStringAddress because of byte size format
;
; 
; Allocate memory
; Set variables
; If possible, load data
; Normalize data
; Set addresses (pointers to data)
; Sort accordingly
; Create Tree / Index
; Perform Lookups
; 
MakeIndex proc hTree :DWORD, Qty :DWORD
    ; ------------------------------------------
    ; Standard Index header for all indexes made
    ;    bottom dword is the record count
    ;    second dword is the entry record number, 1-based
    ; ------------------------------------------
    ; Memory allocation in-relation to index needs is done
    ; outside this procedure and hIndex points to it
    ;
    ; Design spec - What should be maximum size?
    ; (memory blocks of 512, 128K, or what - 16 Gig)
    ;
    ; Design spec - At what point should the 
    ; index be on the harddrive versus using memory?
    ; ------------------------------------------
    ; Qty should be greater than zero
    ;     this procedure does check for this issue
    ; ------------------------------------------
    pushad
    pushfd
    ; ------------------------------------------
         mov edx, hTree                   ; Pointer to memory area for our Index
         mov eax, Qty                     ; Get qty of items to be in the index
         mov [edx], eax                   ; Store qty in the index header
         add edx, 8                       ; We need to account for header area
         .if eax <= 256                   ; ------------------------------------------ Byte format
             lea eax, [edx + eax * 2]     ; Computes top address
             invoke MakeIndex1, edx, eax  ; Creates tree of index offsets
             mov edx, hTree               ; Get Tree address
             sub eax, edx                 ; Removing Base address = index format
             shr eax, 2                   ; Divide by 4 (word adjust)
             mov [edx + 4], eax           ; Store entry point, 1-based into hTree header
         .elseif eax <= 65536             ; ------------------------------------------ Word format
             lea eax, [edx + eax * 2]     ; Computes top address
             invoke MakeIndex2, edx, eax  ; Creates tree of index offsets
             mov edx, hTree               ; Get Tree address
             sub eax, edx                 ; Removing Base address = index format
             shr eax, 3                   ; Divide by 8 (dword adjust)
             mov [edx + 4], eax           ; Store entry point, 1-based into hTree header
         .else                            ; ------------------------------------------ DWORD format
             lea eax, [edx + eax * 2]     ; Computes top address
             invoke MakeIndex4, edx, eax  ; Creates tree of index offsets
             mov edx, hTree               ; Get Tree address
             sub eax, edx                 ; Removing Base address = index format
             shr eax, 4                   ; Divide by 16 (qword adjust)
             mov [edx + 4], eax           ; Store entry point, 1-based in to hTree header
         .endif
    ; ------------------------------------------
    popfd
    popad
    ret
MakeIndex endp
; 
MakeIndex1 proc bottom :DWORD, top :DWORD
    ;------------------------------------------------------------------------------
    ; Re-entrant routine, fills in a word size table with pairs (left - right transverses)
    ; of signed bytes determined by the size of the table.
    ;
    ; While recursing in  - both offsets (left and right) are set to zero
    ; While recursing out - both offsets are set accordingly
    ;
    ; Returns eax = midpoint, which will be the entry point for the lookup routine.
    ;------------------------------------------------------------------------------
    LOCAL midd :DWORD                   ; Holds the address of mid point
    ; ------------------------------------------
        mov edx, bottom                 ; Get bottom value
        mov eax, top                    ; Get top value
        sub eax, edx                    ; Compute range
        shr eax, 2                      ; Divide by 4
        shl eax, 1                      ; Multiply by 2, sets or clears the ZF flag
        lea eax, [eax + edx]            ; Compute the mid point address
        mov midd, eax                   ; Now save it
        mov byte ptr [eax], 0           ; Recursing in we zero our elements
        jz short @F                     ; If ZF flag set then we need to transverse the other side, so jmp
        invoke MakeIndex1, edx, eax     ; Recursive call with new top value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx], al                   ; Save the resulting offset
@@:
        mov edx, midd                   ; Restablish our mid point value
        add edx, 2                      ; Since we are transversing the other half, we need to adjust our mid point
        cmp edx, top                    ; Is this half done yet?
        je short @F                     ; Only if the mid and top are equal, so jmp when they are
        invoke MakeIndex1, edx, top     ; Recursive call with new bottom value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx + 1], al               ; Save the resulting offset
@@:
        mov eax, midd                   ; Return our entry point
    ; ------------------------------------------
    ret
MakeIndex1 endp
; 
MakeIndex2 proc bottom :DWORD, top :DWORD
    ;------------------------------------------------------------------------------
    ; Re-entrant routine, fills in a dword size table with pairs (left - right transverses)
    ; of signed words determined by the size of the table.
    ;
    ; While recursing in  - both offsets (left and right) are set to zero
    ; While recursing out - both offsets are set accordingly
    ;
    ; Returns eax = midpoint, which will be the entry point for the lookup routine.
    ;------------------------------------------------------------------------------
    LOCAL midd :DWORD                   ; Holds the address of mid point
    ; ------------------------------------------
        mov edx, bottom                 ; Get bottom value
        mov eax, top                    ; Get top value
        sub eax, edx
        shr eax, 3                      ; Divide by 8
        shl eax, 2                      ; Muliply by 4, sets or clears the ZF flag
        lea eax, [eax + edx]            ; Compute the mid point address
        mov midd, eax                   ; Now save it
        mov word ptr [eax], 0           ; Recursing in we zero our elements
        jz short @F                     ; If ZF flag then we need to transverse the other side, so jmp
        invoke MakeIndex2, edx, eax     ; Recursive call with new top value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx], ax                   ; Save the resulting offset
@@:
        mov edx, midd                   ; Restablish our mid point value
        add edx, 4                      ; Since we are transversing the other half, we need to adjust our mid point
        cmp edx, top                    ; Is this half done yet?
        je short @F                     ; Only if the two are equal, so jmp when they are
        invoke MakeIndex2, edx, top     ; Recursive call with new bottom value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx + 2], ax               ; Save the resulting offset
@@:
        mov eax, midd                   ; Return our entry point
    ; ------------------------------------------
    ret
MakeIndex2 endp
; 
MakeIndex4 proc bottom :DWORD, top :DWORD
    ;------------------------------------------------------------------------------
    ; Re-entrant routine, fills in a qword size table with pairs (left - right transverses)
    ; of signed dwords determined by the size of the table.
    ;
    ; While recursing in  - both offsets (left and right) are set to zero
    ; While recursing out - both offsets are set accordingly
    ;
    ; Returns eax = midpoint, which will be the entry point for the lookup routine.
    ;------------------------------------------------------------------------------
    LOCAL midd :DWORD                   ; Holds the address of mid point
    ; ------------------------------------------
        mov edx, bottom                 ; Get bottom value
        mov eax, top                    ; Get top value
        sub eax, edx
        shr eax, 4                      ; Divide by 16
        shl eax, 3                      ; Multiply by 8, sets or clears the ZF flag
        lea eax, [eax + edx]            ; Compute the mid point address
        mov midd, eax                   ; Now save it
        mov dword ptr [eax], 0          ; Recursing in we zero our elements
        jz short @F                     ; If ZF flag is set then we need to transverse the other side, so jmp
        invoke MakeIndex4, edx, eax     ; Recursive call with new top value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx], eax                  ; Save the resulting offset
@@:
        mov edx, midd                   ; Restablish our mid point value
        add edx, 8                      ; Since we are transversing the other half, we need to adjust our mid point
        cmp edx, top                    ; Is this half done yet?
        je short @F                     ; Only if the two are equal, so jmp when they are
        invoke MakeIndex4, edx, top     ; Recursive call with new bottom value
        mov edx, midd                   ; Restablish our mid point value
        sub eax, edx                    ; Compute our offset
        mov [edx + 4], eax              ; Save the resulting offset
@@:
        mov eax, midd                   ; Return our entry point
    ; ------------------------------------------
    ret
MakeIndex4 endp
; 
IndexedLookup proc item :DWORD, hTree :DWORD, hAddr :DWORD, comparer :NEAR, vTerm :WORD
    push esi
    push edi
    push ebx
    push ecx
    push edx
    ; ------------------------------------------
         mov esi, hTree                   ; Pointer to our tree
         mov eax, [esi]                   ; Get qty from the index header
         .if eax <= 256
             invoke IndexedLookup1, item, hAddr, hTree, comparer, vTerm
         .elseif eax <= 65536
             invoke IndexedLookup2, item, hAddr, hTree, comparer, vTerm
         .else
             invoke IndexedLookup4, item, hAddr, hTree, comparer, vTerm
         .endif
    ; ------------------------------------------
    pop edx
    pop ecx
    pop ebx
    pop edi
    pop esi
    ret
IndexedLookup endp
; 
IndexedLookup1 proc item :DWORD, hAddr :DWORD, hTree :DWORD, comparer :NEAR, vTerm :WORD
    ; -------------------------------------------------------------------
    ; returns:
    ;   NZ and eax = the 1-based record number of the item if it is found
    ;   ZF and eax = 0 otherwise.
    ; -------------------------------------------------------------------
    push esi
    push edi
    push ebx
    ; -------------------------------------------------------------------
         mov edi, hAddr            ; Get location of addresses
         mov esi, hTree            ; Get our tree index
         mov ebx, [esi + 4]        ; Get entry Index number, 1-based
         add esi, 8                ; Bypass Header area
IL_loop:
         push vTerm                ; Terminator value
         push [edi + ebx * 4]      ; Push address to item location
         push item                 ; Push address of search item location
         call comparer             ; Test the two strings
         jz short IL_success       ; If we found the correct string then exit
         jc short IL_look_higher   ; Determine which way to search again
IL_look_lower:
         movsx eax, BYTE PTR [esi + ebx * 1 + 1] ; Get offset
         jmp short IL_adjust
IL_look_higher:
         movsx eax, BYTE PTR [esi + ebx * 1] ; Get offset
IL_adjust:
         test eax, eax             ; Is offset = 0
         jz short IL_done          ; If so jmp
         add ebx, eax              ; else adjust index by offset
         jmp short IL_loop         ; Repeat search
IL_success:
         xchg eax, ebx             ; Put our index in eax
         inc eax                   ; Convert to 1-based number
IL_done:
    ; -------------------------------------------------------------------
    pop ebx
    pop edi
    pop esi
    ret
IndexedLookup1 endp
; 
IndexedLookup2 proc item :DWORD, hAddr :DWORD, hTree :DWORD, comparer :NEAR, vTerm :WORD
    ; -------------------------------------------------------------------
    ; returns:
    ;   NZ and eax = the 1-based record number of the item if it is found
    ;   ZF and eax = 0 otherwise.
    ; -------------------------------------------------------------------
    push esi
    push edi
    push ebx
    ; -------------------------------------------------------------------
         mov edi, hAddr            ; Get location of addresses
         mov esi, hTree            ; Get our tree index
         mov ebx, [esi + 4]        ; Get entry Index number, 1-based
         add esi, 8                ; Bypass Header area
IL_loop:
         push vTerm                ; Terminator value
         push [edi + ebx * 4]      ; Push address to item location
         push item                 ; Push address of search item location
         call comparer             ; Test the two strings
         jz short IL_success       ; If we found the correct string then exit
         jc short IL_look_higher   ; Determine which way to search again
IL_look_lower:
         movsx eax, WORD PTR [esi + ebx * 2 + 2] ; Get offset
         jmp short IL_adjust
IL_look_higher:
         movsx eax, WORD PTR [esi + ebx * 2] ; Get offset
IL_adjust:
         test eax, eax             ; Is offset = 0
         jz short IL_done          ; If so jmp
         add ebx, eax              ; else adjust index by offset
         jmp short IL_loop         ; Repeat search
IL_success:
         xchg eax, ebx             ; Put our index in eax
         inc eax                   ; Convert to 1-based number
IL_done:
    ; -------------------------------------------------------------------
    pop ebx
    pop edi
    pop esi
    ret
IndexedLookup2 endp
; 
IndexedLookup4 proc item :dword, hAddr :dword, hTree :dword, comparer :near, vTerm :WORD
    ; -------------------------------------------------------------------
    ; returns:
    ;   NZ and eax = the 1-based record number of the item if it is found
    ;   ZF and eax = 0 otherwise.
    ; -------------------------------------------------------------------
    push esi
    push edi
    push ebx
    ; -------------------------------------------------------------------
         mov edi, hAddr            ; Get location of addresses
         mov esi, hTree            ; Get our tree index
         mov ebx, [esi + 4]        ; Get entry Index number, 1-based
         add esi, 8                ; Bypass Header area
IL_loop:
         push vTerm                ; Terminator value
         push [edi + ebx * 4]      ; Push address to item location
         push item                 ; Push address of search item location
         call comparer             ; Test the two strings
         jz short IL_success       ; If we found the correct string then exit
         jc short IL_look_higher   ; Determine which way to search again
IL_look_lower:
         mov eax, [esi + ebx * 4 + 4] ; Get offset
         jmp short IL_adjust
IL_look_higher:
         mov eax, [esi + ebx * 4]  ; Get offset
IL_adjust:
         test eax, eax             ; Is offset = 0
         jz short IL_done          ; If so jmp
         add ebx, eax              ; else adjust index by offset
         jmp short IL_loop         ; Repeat search
IL_success:
         xchg eax, ebx             ; Put our index in eax
         inc eax                   ; Convert to 1-based number
IL_done:
    ; -------------------------------------------------------------------
    pop ebx
    pop edi
    pop esi
    ret
IndexedLookup4 endp
; 
line_cmp proc ptr1 :DWORD, ptr2 :DWORD, vTerm :WORD
    ;------------------------------------------------------------------------------
    ; case-sensitive, and returns info in the flags only.
    ;------------------------------------------------------------------------------
    push ebx
    push ecx
    push edx
    push esi
    push edi
    ; ------------------------------------------
         mov esi, ptr1             ; Get search string address
         mov edi, ptr2             ; Get lookup key string address
@@:
         lodsb                     ; Get char from string1
         mov ah, [edi]             ; Get char from string2
         inc edi                   ; For sync between lodsb and procedure
         cmp ax, vTerm             ; Test to see if string1 = string2
         je short @F               ; If so, exit
         cmp al, ah                ; Test char1 = char2
         jz short @B               ; If so, repeat compare chars
@@:
    ; ------------------------------------------
    pop edi
    pop esi
    pop edx
    pop ecx
    pop ebx
    ret
line_cmp endp
; 
line_cmp_nocase proc ptr1 :DWORD, ptr2 :DWORD, vTerm :WORD
    ;------------------------------------------------------------------------------
    ; case-sensitive, and returns info in the flags only.
    ;------------------------------------------------------------------------------
    push ebx
    push ecx
    push edx
    push esi
    push edi
    ; ------------------------------------------
         mov esi, ptr1             ; Get string1 address
         mov edi, ptr2             ; Get string2 address
@@:
         lodsb                     ; Get char from string1
         mov ah, [edi]             ; Get char from string2
         inc edi                   ; For sync between lodsb and procedure
         cmp ax, vTerm             ; Test to see if string1 = string2
         je short @F               ; If so, exit
         capal                     ; Convert char1 to Capital format
         xchg al, ah               ; Exchange characters
         capal                     ; Convert char2 to Capital format
         xchg al, ah               ; Exchange characters
         cmp al, ah                ; Test char1 = char2
         jz short @B               ; If so, repeat compare chars
@@:
    ; ------------------------------------------
    pop edi
    pop esi
    pop edx
    pop ecx
    pop ebx
    ret                            ; If not, flags have condition of compare
line_cmp_nocase endp
; 
line_cmp_numb1 proc ptr1 :DWORD, ptr2 :DWORD, vTerm :WORD
    push esi
    push edi
    push eax
    ; --------------------
         mov esi, ptr1             ; Get address
         mov edi, ptr2             ; Get address
         movzx eax, BYTE PTR [esi] ; Get value
         mov esi, eax              ; Transfer control
         movzx eax, BYTE PTR [edi] ; Get value
         mov edi, eax              ; Transfer control
         cmp esi, edi              ; Compare the values, set flags accordingly
    ; --------------------
    pop eax
    pop edi
    pop esi
    ret
line_cmp_numb1 endp
; 
line_cmp_numb2 proc ptr1 :DWORD, ptr2 :DWORD, vTerm :WORD
    push esi
    push edi
    push eax
    ; --------------------
         mov esi, ptr1             ; Get address
         mov edi, ptr2             ; Get address
         movzx eax, WORD PTR [esi] ; Get value
         mov esi, eax              ; Transfer control
         movzx eax, WORD PTR [edi] ; Get value
         mov edi, eax              ; Transfer control
         cmp esi, edi              ; Compare the values, set flags accordingly
    ; --------------------
    pop eax
    pop edi
    pop esi
    ret
line_cmp_numb2 endp
; 
line_cmp_numb4 proc ptr1 :DWORD, ptr2 :DWORD, vTerm :WORD
    push esi
    push edi
    ; --------------------
         mov esi, ptr1        ; Get address
         mov edi, ptr2        ; Get address
         mov esi, [esi]       ; Get value
         mov edi, [edi]       ; Get value
         cmp esi, edi         ; Compare the values, set flags accordingly
    ; --------------------
    pop edi
    pop esi
    ret
line_cmp_numb4 endp
; 
SetStringAddresses proc hAddr :DWORD, hStrings :DWORD, linecount :DWORD, terminator :BYTE
    pushad
    pushfd
    ; ------------------------------------------
        mov ebx, linecount         ; This is how many strings we have
        mov edx, hAddr             ; Get handle to address table
        mov edi, hStrings          ; Get handle to string table or numeric table
        mov al, terminator         ; Get terminator value
        or ebx, -1                 ; Initialize counter
        jmp Jmp_In                 ; Perform procedure
    Scan_Strings:
        mov [edx + ebx * 4], edi   ; Base + Index = pointer
        repne scasb                ; Check string characters until terminator value found, edi and ecx autochanges
    Jmp_In:
        or ecx, -1                 ; Initialize counter - make sure its not zero (for repne above)
        inc ebx                    ; Increment count
        cmp ebx, linecount         ; Are we done yet?
        jne short Scan_Strings     ; If not, then repeat
    ; ------------------------------------------
    popfd
    popad
    ret
SetStringAddresses endp
; 
